class Canvas:
    def __init__(self, screen):
        self._screen = screen

    def draw_text(self, text: str, position: tuple):
        """
        """
        x, y = position
        try:
            self._screen.addstr(y, x, text)
        except Exception:
            pass

    def draw_char(self, char: str, position: tuple):
        """
        """
        x, y = position
        try:
            self._screen.addnstr(y, x, char, 1)
        except Exception:
            pass

    def draw_frame(self, frame: tuple, style=None, thick=False, dashed=False):
        """
        Draw a rectangular frame.

        :param frame: tuple of (x, y, width, height)
        """
        # TODO move definitions to separate module
        # TODO make unicode configurable or better: dependent on unicode
        # support in terminal
        UNICODE = True
        if UNICODE:
            if thick:
                VLINE = '\u2588'
                UHLINE = '\u2580'
                LHLINE = '\u2584'
                ULCORNER = '\u2588'
                URCORNER = '\u2588'
                LLCORNER = '\u2588'
                LRCORNER = '\u2588'
            else:
                VLINE = '\u2502'
                UHLINE = '\u2500'
                LHLINE = '\u2500'
                ULCORNER = '\u250C'
                URCORNER = '\u2510'
                LLCORNER = '\u2514'
                LRCORNER = '\u2518'
        else:
            if thick:
                VLINE = '#'
                UHLINE = '#'
                LHLINE = '#'
                ULCORNER = '#'
                URCORNER = '#'
                LLCORNER = '#'
                LRCORNER = '#'
            else:
                VLINE = '|'
                UHLINE = '-'
                LHLINE = '_'
                ULCORNER = '.'
                URCORNER = '.'
                LLCORNER = '|'
                LRCORNER = '|'
        x, y, w, h = frame
        for by in range(1, h - 1):
            if dashed and by % 2 == 1:
                continue
            self.draw_char(VLINE, (x, y + by))
            self.draw_char(VLINE, (x + w - 1, y + by))
        for bx in range(1, w - 1):
            if dashed and bx % 2 == 1:
                continue
            self.draw_char(UHLINE, (x + bx, y))
            self.draw_char(LHLINE, (x + bx, y + h - 1))
        self.draw_char(ULCORNER, (x, y))
        self.draw_char(URCORNER, (x + w - 1, y))
        self.draw_char(LLCORNER, (x, y + h - 1))
        self.draw_char(LRCORNER, (x + w - 1, y + h - 1))

    def draw_box(self, box: tuple, style=None):
        """
        Draw a rectangular box.

        :param box: tuple of (x, y, width, height)
        """
        x, y, w, h = box
        UNICODE = True
        BOX = '\u2588' if UNICODE else '#'
        for by in range(h):
            for bx in range(w):
                self.draw_char(BOX, (x + bx, y + by))



from ..events import EventNode
from ..styles import WidgetStyle


class Widget(EventNode):
    def __init__(self, parent=None, address=''):
        super().__init__(parent, address)
        self._style = WidgetStyle()
        self._box = (0, 0, -1, -1)
        self._focusable = False
        self._focused = False
        self._visible = True

    def _handle_key(self, key):
        return key

    def handle_key(self, key):
        return self._handle_key(key)

    def pack(self):
        return

    def _show(self, canvas):
        return

    def show(self, canvas):
        self._show(canvas)



from dataclasses import dataclass
from queue import Queue
from typing import Any, Optional


_callbacks = {}
_queue = Queue()


def send_event(event_type, data=None, source=None, sink=None):
    """
    Throws a new event onto the event bus. Events can be sent from any source,
    but they can only be received by EventNodes.

    :param event_type:
    :param data: event payload
    :param source: where the event comes from
    :param sink: event destination
    """
    event = None
    if isinstance(event_type, Event):
        event = event_type
    else:
        event = Event(event_type, data, source, sink)
    _queue.put(event)


def dispatch_events():
    """
    Distribute all events in the queue among the listening receivers.
    Called inside the mainloop.
    """
    while not _queue.empty():
        event = _queue.get()
        if event.type in _callbacks:
            # directed events
            if event.sink:
                if event.sink in _callbacks[event.type]:
                    for callback in _callbacks[event.type][event.sink]:
                        callback(event)
            # broadcast events
            else:
                for address in _callbacks[event.type]:
                    for callback in _callbacks[event.type][address]:
                        callback(event)


@dataclass(frozen=True)
class Event:
    type: str
    data: Optional[Any]
    source: Optional[str]
    sink: Optional[str]


class EventNode:
    """
    EventNode objects process events sent from arbitrary sources.

    :param address: address string
    """
    _counter = 0

    def __init__(self, parent=None, address=''):
        self.parent = parent
        self.address = address
        type(self)._counter += 1
        if not address:
            self.address = ''
            if parent:
                self.address = parent.address + '.'
            classname = self.__class__.__name__
            self.address += '{}:{}'.format(classname, type(self)._counter)
        # register all '@on'-decorated methods
        for member in dir(self):
            if callable(getattr(self, member)):
                func = getattr(self, member)
                if hasattr(func, 'register'):
                    _register_callback(self, func.event_type, func)

    def register_callback(self, event_type, func):
        _register_callback(self, event_type, func)

    def send_event(self, event_type, data=None, sink=None):
        """
        Wrapper method for send_event in event module.
        Reduces number of parameters needed by inserting source address.

        :param event_type:
        :param data: event payload
        :param sink: event destination
        """
        send_event(event_type, data, self.address, sink)

    def __str__(self):
        return '<EventNode address={}>'.format(self.address)


def _register_callback(node: EventNode, event_type, func):
    """
    :param node: EventNode object
    :param event_type:
    :param func: callback method of node
    """
    if event_type not in _callbacks:
        _callbacks[event_type] = {}
    if node.address not in _callbacks[event_type]:
        _callbacks[event_type][node.address] = []
    _callbacks[event_type][node.address].append(func)


def on(event_type):
    """
    Decorator for EventNodes.
    Subscribes a method to an event type.

    Example::

        class Listener(EventNode):
            ...
            @on('listener.hello')
            def hello(self, event):
                self.do_something(event.data)
            ...

    :param event_type:
    """
    def decorator(func):
        from functools import wraps

        @wraps(func)
        def decorate(*args, **kwargs):
            return func(*args, **kwargs)

        decorate.register = True
        decorate.event_type = event_type
        return decorate
    return decorator



from stewie.application import Application


def demo_vbox():
    widgettree = {
        'type': 'VBox',
        'children': [
            {
                'type': 'Label',
                'text': ''
            },
            {
                'type': 'Button',
                'text': 'button'
            },
            {
                'type': 'OptionBox',
                'options': ['foo', 'bar', 'baz']
            },
            {
                'type': 'CheckBox',
                'text': 'boolean'
            },
            {
                'type': 'ProgressBar',
                'progress': 0.15
            }
        ]
    }
    app = Application(widgettree)
    def activate(event):
        label = app.frame.get_child('Label:1')
        source = app.frame.get_child(event.source)
        label.set_text(event.source)
        if event.source == 'Button:1':
            label.set_text('button pressed')
        elif event.source == 'OptionBox:1':
            label.set_text('optionbox activated: ' + source.get_option())
        elif event.source == 'CheckBox:1':
            label.set_text('checkbox toggled: ' + str(source._state))
    app.register_callback('activate', activate)
    app.run()



from .widget import Widget
from ..utils import strsize


class CheckBox(Widget):
    def __init__(self, text: str, state=False, parent=None, address=''):
        super().__init__(parent, address)
        self._text = text
        self._state = state
        self._focusable = True

    def _show(self, canvas):
        x, y, w, h = self._box
        if self._focused:
            canvas.draw_frame(self._box, thick=True)
        else:
            canvas.draw_frame(self._box, thick=False, dashed=True)
        if self._text:
            text = '['
            text += 'x' if self._state else ' '
            text += '] ' + self._text
            textw, texth = strsize(text)
            textx = int((w - textw) / 2)
            texty = int((h - texth) / 2)
            canvas.draw_text(text, (x + textx, y + texty))

    def _handle_key(self, key):
        if key == ord(' '):
            self._state = not self._state
            self.send_event('activate', data=self._state)
            return
        return key



class BoxAttribute:
    def __init__(self, top=None, bottom=None, left=None, right=None):
        self._top = top
        self._bottom = bottom
        self._left = left
        self._right = right

    @property
    def top(self):
        return self._top

    @top.setter
    def top(self, t):
        self._top = t

    @property
    def bottom(self):
        return self._bottom

    @bottom.setter
    def bottom(self, b):
        self._bottom = b

    @property
    def left(self):
        return self._left

    @left.setter
    def left(self, l):
        self._left = l

    @property
    def right(self):
        return self._right

    @right.setter
    def right(self, r):
        self._right = r



from .boxattribute import BoxAttribute
from .sizeattribute import SizeAttribute


class WidgetStyle:
    def __init__(self):
        self._padding = BoxAttribute(0, 0, 0, 0)
        self._margin = BoxAttribute(0, 0, 0, 0)
        self._width = SizeAttribute()
        self._height = SizeAttribute()

    @property
    def padding(self):
        return self._padding

    @padding.setter
    def padding(self, p):
        if type(p) == BoxAttribute:
            self._padding = p
        else:
            raise RuntimeError('wrong type passed for padding ' + str(type(p)))

    @property
    def margin(self):
        return self._margin

    @margin.setter
    def marging(self, m):
        if type(m) == BoxAttribute:
            self._margin = m
        else:
            raise RuntimeError('wrong type passed for margin ' + str(type(m)))


class ContainerStyle(WidgetStyle):
    def __init__(self):
        super(WidgetStyle).__init__()



from .container import Container
from .frame import Frame
from .hbox import HBox
from .listbox import ListBox
from .vbox import VBox

__all__ = [
    'Container',
    'Frame',
    'HBox',
    'ListBox',
    'VBox'
]



from .container import Container


class Frame(Container):
    def __init__(self, box, parent, address='frame'):
        super().__init__(parent, address)
        self._box = box

    def add_child(self, widget):
        if len(self._children) > 0:
            raise RuntimeError('Frames can only have one child')
        super().add_child(widget)

    def _pack(self):
        if self._children:
            self._children[0]._box = self._box



import argparse

from .listbox import demo_listbox
from .vbox import demo_vbox


demos = {
    'listbox': demo_listbox,
    'vbox': demo_vbox
}


def main():
    parser = argparse.ArgumentParser()
    choices = list(demos.keys())
    parser.add_argument('DEMO',
                        choices=choices)
    args = parser.parse_args()
    demo = args.DEMO
    demos.get(demo)()



import curses

from .container import Container


class HBox(Container):
    def __init__(self, parent=None, address=''):
        super().__init__(parent, address)
        self._keys = {
            'left': curses.KEY_LEFT,
            'right': curses.KEY_RIGHT
        }

    def _pack(self):
        x, y, w, h = self._box
        if not self._children:
            return
        childwidth = w / len(self._children)
        for c in range(len(self._children)):
            cx = int(c * childwidth)
            cy = y
            cw = int(childwidth)
            ch = h
            self._children[c]._box = (cx, cy, cw, ch)

    def _handle_key(self, key):
        indices = self._get_focusable_child_indices()
        index = indices.index(self._focused_child)
        if key == self._keys.get('right', curses.KEY_RIGHT):
            index += 1
            if index >= len(indices):
                index = 0
            self._focused_child = indices[index]
            return
        if key == self._keys.get('left', curses.KEY_LEFT):
            index -= 1
            if index < 0:
                index = len(indices) - 1
            self._focused_child = indices[index]
            return
        return key



import curses


def curses_keycode(s):
    if s.isalnum():
        return ord(s)
    keys = {
        'up': curses.KEY_UP,
        'down': curses.KEY_DOWN,
        'left': curses.KEY_LEFT,
        'right': curses.KEY_RIGHT,
        'space': ord(' ')
    }
    return keys.get(s, None)



from .widget import Widget


class NumberBox(Widget):
    def __init__(self, text: str, parent=None, address=''):
        super().__init__(parent, address)
        self._number = 0
        self._focusable = True

    def _show(self, canvas):
        x, y, w, h = self._box
        if self._focused:
            canvas.draw_frame(self._box, thick=True)
        else:
            canvas.draw_frame(self._box, thick=False, dashed=True)
        text = str(self._number)
        textw, texth = strsize(text)
        textx = w - textw - 1
        texty = int((h - texth) / 2)
        canvas.draw_text(text, (x + textx, y + texty))

    def _handle_key(self, key):
        if ord('0') <= key <= ord('9'):
            n = int(key)
            self._number
            return
        return key



from .boxattribute import BoxAttribute
from .sizeattribute import SizeAttribute


class BoxStyle:
    def __init__(self):
        self._width = SizeAttribute()
        self._height = SizeAttribute()
        self._border = BoxAttribute()



class SizeAttribute:
    def __init__(self, min_size=1, max_size=-1, prefered_size=-1):
        self._min_size = min_size
        self._max_size = max_size
        self._prefered_size = prefered_size
        if self._prefered_size == -1:
            self._prefered_size = max_size



def ellipsis(string, length):
    if not string:
        return ''
    if not isinstance(string, str):
        string = str(string)
    if len(string) <= length:
        return string
    if length <= 3:
        return string[:length-1] + '…'
    return string[:length-3] + '...'


def padstr(string, length, center=False):
    delta = length - len(string)
    if delta > 0:
        if center:
            prefix = delta // 2
            suffix = delta - prefix
            string = ' ' * prefix + string + ' ' * suffix
        else:
            string = string + ' ' * delta
    return string


def strsize(string):
    height = string.count('\n') + 1
    width = 0
    for line in string.splitlines():
        if len(line) > width:
            width = len(line)
    return width, height



import curses

from .container import Container


class ListBox(Container):
    def __init__(self, visible_entries: int, parent=None, address=''):
        super().__init__(parent, address)
        self._visible_entries = visible_entries
        self._scroll = 0
        self._keys = {
            'up': curses.KEY_UP,
            'down': curses.KEY_DOWN
        }

    def _add_child(self, child):
        if not child._focusable:
            raise RuntimeError('ListBox children must be focusable')
        return True

    def _remove_child(self, child):
        return True

    def _pack(self):
        if not self._children:
            return
        x, y, w, h = self._box
        childheight = int(h / self._visible_entries)
        for child in self._children:
            child._visible = False
        start = self._scroll
        end = start + self._visible_entries
        if end > len(self._children):
            end = len(self._children)
        for c in range(start, end):
            cx = x
            cy = int((c - start) * childheight)
            cw = w
            ch = int(childheight)
            self._children[c]._visible = True
            self._children[c]._box = (cx, cy, cw, ch)

    def _show(self, canvas):
        return

    def show(self, canvas):
        """
        Override Container's show method so that only a portion of the
        children is displayed.
        """
        self._show(canvas)
        start = self._scroll
        end = start + self._visible_entries
        if end > len(self._children):
            end = len(self._children)
        for c in range(start, end):
            child = self._children[c]
            if child._visible:
                child.show(canvas)

    def _handle_key(self, key):
        if key == self._keys.get('down', curses.KEY_DOWN):
            self._focused_child += 1
            if self._focused_child >= len(self._children):
                self._focused_child = len(self._children) - 1
            key = None
        elif key == self._keys.get('up', curses.KEY_UP):
            self._focused_child -= 1
            if self._focused_child < 0:
                self._focused_child = 0
            key = None
        if self._focused_child - self._scroll >= self._visible_entries:
            self._scroll += self._visible_entries
        elif self._focused_child - self._scroll < 0:
            self._scroll -= self._visible_entries
        if key is None:
            self.pack()
        return key



from .widget import Widget
from ..utils import strsize


class OptionBox(Widget):
    def __init__(self, options: list, parent=None, address=''):
        super().__init__(parent, address)
        self._options = options
        self._ptr = 0
        self._focusable = True
        self._keys = {
            'activate': ord(' '),
            'previous': ord('-'),
            'next': ord('+')
        }

    def get_option(self):
        return self._options[self._ptr]

    def _show(self, canvas):
        x, y, w, h = self._box
        if self._focused:
            canvas.draw_frame(self._box, thick=True)
        else:
            canvas.draw_frame(self._box, thick=False, dashed=True)
        if self._options:
            option = self._options[self._ptr]
            textw, texth = strsize(option)
            textx = int((w - textw) / 2)
            texty = int((h - texth) / 2)
            canvas.draw_text(option, (x + textx, y + texty))

    def _handle_key(self, key):
        if key == self._keys.get('previous', ord('-')):
            if self._ptr - 1 >= 0:
                self._ptr -= 1
                self.send_event('change', data=self._options[self._ptr])
            return
        elif key == self._keys.get('next', ord('+')):
            if self._ptr + 1 < len(self._options):
                self._ptr += 1
                self.send_event('change', data=self._options[self._ptr])
            return
        elif key == self._keys.get('activate', ord(' ')):
            self.send_event('activate')
            return
        return key



from ..styles import ContainerStyle
from ..widgets import Widget


class Container(Widget):
    def __init__(self, parent=None, address=''):
        super().__init__(parent, address)
        self._focused_child = 0
        self._focusable = True
        self._children = []
        self._style = ContainerStyle()

    def get_child(self, address, recursive=True):
        for child in self._children:
            if child.address == address:
                return child
            if recursive:
                if hasattr(child, 'get_child'):
                    c = child.get_child(address, recursive)
                    if c is not None:
                        return c
        return None

    def _add_child(self, widget):
        return True

    def add_child(self, widget):
        if self._add_child(widget):
            widget._parent = self
            self._children.append(widget)
            self.pack()
            return True
        return False

    def _remove_child(self, widget):
        return True

    def remove_child(self, widget):
        status = False
        for child in self._children:
            if child.address == widget.address:
                if self._remove_child(child):
                    self._children.remove(child)
                    status = True
                    break
        if status:
            self.pack()
        return status

    def _get_focusable_child_indices(self):
        indices = []
        for c in range(len(self._children)):
            child = self._children[c]
            if child._focusable:
                indices.append(c)
        return indices

    def _pack(self):
        return

    def pack(self):
        if self._focused_child >= len(self._children):
            self._focused_child = len(self._children) - 1
        self._pack()
        for c in range(len(self._children)):
            child = self._children[c]
            child._focused = (self._focused and c == self._focused_child)
            child.pack()

    def _show(self, canvas):
        return

    def show(self, canvas):
        if not self._visible:
            return
        self._show(canvas)
        for child in self._children:
            if child._visible:
                child.show(canvas)

    def _handle_key(self, key):
        return key

    def handle_key(self, key):
        k = self._handle_key(key)
        for c in range(len(self._children)):
            child = self._children[c]
            child._focused = (c == self._focused_child)
        if k is None:
            return
        if 0 <= self._focused_child < len(self._children):
            child = self._children[self._focused_child]
            if child.handle_key(key) is None:
                return
        return key

    def __iter__(self):
        return self._children.__iter__()

    def __len__(self):
        return len(self._children)



from .boxattribute import BoxAttribute
from .boxstyle import BoxStyle
from .sizeattribute import SizeAttribute
from .style import ContainerStyle, WidgetStyle


__all__ = [
    'BoxAttribute',
    'BoxStyle',
    'SizeAttribute',
    'ContainerStyle',
    'WidgetStyle'
]



from .widget import Widget
from ..utils import strsize


class ProgressBar(Widget):
    _progress = 0

    def __init__(self, progress=0, parent=None, address=''):
        super().__init__(parent, address)
        self.progress = progress

    @property
    def progress(self):
        return self._progress

    @progress.setter
    def progress(self, progress: float):
        if 0 <= progress <= 1:
            self._progress = progress
        else:
            raise RuntimeError('progress value not in interval [0, 1]')

    def show(self, canvas):
        x, y, w, h = self._box
        # TODO vertical/horizontal option
        # TODO reverse option
        w_fill = int(w * self.progress)
        canvas.draw_box((x, y, w_fill, h))
        canvas.draw_frame(self._box)
        ptext = '{}%'.format(int(self.progress * 100))
        textw, texth = strsize(ptext)
        textx = int((w - textw) / 2)
        texty = int((h - texth) / 2)
        canvas.draw_text(ptext, (x + textx, y + texty))



class ColorAttribute:
    def __init__(self):
        self._foreground = 1
        self._background = 0
        self._pattern = '#'

    @property
    def foreground(self):
        return self._foreground

    @foreground.setter
    def foreground(self, f):
        self._foreground = f

    @property
    def background(self):
        return self._background

    @background.setter
    def background(self, b):
        self._background = b

    @property
    def pattern(self):
        return self._pattern

    @pattern.setter
    def pattern(self, p):
        self._pattern = p



import curses

import traceback
import signal
import sys

from .canvas import Canvas
from .containers import Frame
from .events import dispatch_events, EventNode
from .keyboard import curses_keycode
from .widgettree import build_widget_tree


class Application(EventNode):
    def __init__(self, widgets: dict):
        super().__init__(None, 'application')
        signal.signal(signal.SIGINT, self._exit_signal_handler)
        signal.signal(signal.SIGTERM, self._exit_signal_handler)

        self._screen = curses.initscr()
        try:
            self._canvas = Canvas(self._screen)
            self.frame = Frame((0, 0, curses.COLS, curses.LINES), self)
            build_widget_tree(self.frame, widgets)
        except Exception:
            self.exit()
            exc_info = sys.exc_info()
            traceback.print_exception(*exc_info)
            exit()

    def _exit_signal_handler(self, signum, frame):
        self.exit()
        exit()

    def run(self, thread=None):
        try:
            self.frame.pack()
            curses.noecho()
            curses.cbreak()
            curses.curs_set(False)
            self._screen.keypad(True)
            self._screen.timeout(0)
            height, width = self._screen.getmaxyx()
            quit = False
            while not quit:
                if thread and thread.is_stop_requested():
                    quit = True
                lines, cols = self._screen.getmaxyx()
                if width != cols or height != lines:
                    width = cols
                    height = lines
                    self.frame._box = (0, 0, cols, lines)
                    self.frame.pack()
                self._screen.erase()
                self.frame.show(self._canvas)
                c = self._screen.getch()
                if c == curses_keycode('q'):
                    quit = True
                if c != -1:
                    self.frame.handle_key(c)
                dispatch_events()
            self.exit()
        except Exception:
            self.exit()
            exc_info = sys.exc_info()
            traceback.print_exception(*exc_info)

    def exit(self):
        curses.curs_set(True)
        curses.nocbreak()
        curses.echo()
        self._screen.keypad(False)
        curses.endwin()



from .containers import *
from .widgets import *


# TODO automatically associate widget classes with their names
name2widget = {
    'CheckBox': CheckBox,
    'Frame': Frame,
    'HBox': HBox,
    'ListBox': ListBox,
    'VBox': VBox,
    'Button': Button,
    'Label': Label,
    'OptionBox': OptionBox,
    'ProgressBar': ProgressBar
}


def _walk_children(node, nodewidget):
    children = node.get('children', [])
    for child in children:
        options = {
            k: child[k] for k in child if k not in ('children', 'type')
        }
        try:
            childtype = child['type']
        except Exception:
            raise RuntimeError('child widget type has to be defined')
        childwidget = name2widget.get(childtype)(**options)
        _walk_children(child, childwidget)
        nodewidget.add_child(childwidget)


def build_widget_tree(parent: Container, widgets: dict):
    """
    :param parent: parent Container to which the widgets are added as children
    :param widgets: dictionary representing the widget tree
    """
    root = widgets
    try:
        roottype = root['type']
    except Exception:
        raise RuntimeError('root widget type has to be defined')
    options = {
        k: root[k] for k in root if k not in ('children', 'type')
    }
    rootwidget = name2widget.get(roottype)(**options)
    _walk_children(root, rootwidget)
    parent.add_child(rootwidget)



import curses

from .container import Container


class VBox(Container):
    def __init__(self, parent=None, address=''):
        super().__init__(parent, address)
        self._keys = {
            'up': curses.KEY_UP,
            'down': curses.KEY_DOWN
        }

    def _pack(self):
        x, y, w, h = self._box
        if not self._children:
            return
        childheight = int(h / len(self._children))
        for c in range(len(self._children)):
            cx = x
            cy = y + int(c * childheight)
            cw = w
            ch = int(childheight)
            self._children[c]._box = (cx, cy, cw, ch)
        indices = self._get_focusable_child_indices()
        if self._focused_child in indices:
            index = indices.index(self._focused_child)
        else:
            index = 0
        if len(indices) > 0:
            self._focused_child = indices[index]

    def _show(self, canvas):
        return

    def _handle_key(self, key):
        if self._focused_child < 0:
            return key
        indices = self._get_focusable_child_indices()
        index = indices.index(self._focused_child)
        if key == self._keys.get('down', curses.KEY_DOWN):
            index += 1
            if index >= len(indices):
                index = 0
            self._focused_child = indices[index]
            self.pack()
            return
        if key == self._keys.get('up', curses.KEY_UP):
            index -= 1
            if index < 0:
                index = len(indices) - 1
            self._focused_child = indices[index]
            self.pack()
            return
        return key



#!/usr/bin/env python3
"""
./setup.py file using setuptools.

Includes flake8 command support.
"""
import os

from setuptools import find_packages, setup

from sphinx.setup_command import BuildDoc


NAME = 'stewie'

lvars = {}
PROJECT_DIR = os.path.split(__file__)[0]
if not PROJECT_DIR:
    PROJECT_DIR = '.'
version_py = '{}/{}/__version__.py'.format(PROJECT_DIR, NAME)
with open(version_py, 'r') as f:
    code = compile(f.read(), version_py, 'exec')
    exec(code, {}, lvars)
VERSION = lvars.get('__version__')
RELEASE = lvars.get('__release__')

with open('README.md', 'r') as f:
    readme = f.read()


setup(
    name=NAME,
    author='Henry Krumb',
    author_email='henry.krumb@computerwerk.org',
    version=VERSION,
    url='https://github.com/henrykrumb/stewie',
    description='Simple Terminal Widget Engine',
    long_description=readme,
    packages=find_packages(),
    include_package_data=True,
    cmdclass={
        'build_sphinx': BuildDoc
    },
    command_options={
        'build_sphinx': {
            'project': ('setup.py', NAME),
            'version': ('setup.py', VERSION),
            'release': ('setup.py', RELEASE)
        }
    },
    setup_requires=[
        'docutils',
        'flake8',
        'sphinx'
    ]
)



from .widget import Widget
from ..utils import strsize


class Label(Widget):
    def __init__(self, text: str, parent=None, address=''):
        super().__init__(parent, address)
        self._text = text

    def set_text(self, text):
        self._text = text
        self.pack()

    def _show(self, canvas):
        x, y, w, h = self._box
        if self._text:
            textw, texth = strsize(self._text)
            textx = round((w - textw) / 2)
            texty = round((h - texth) / 2)
            canvas.draw_text(self._text, (x + textx, y + texty))



from .button import Button
from .checkbox import CheckBox
from .label import Label
from .optionbox import OptionBox
from .progressbar import ProgressBar
from .widget import Widget


__all__ = [
    'Button',
    'CheckBox',
    'Label',
    'OptionBox',
    'ProgressBar',
    'Widget'
]



from stewie.application import Application
from stewie.widgettree import build_widget_tree


def make_list_item(parent, i):
    item = {
        'type': 'VBox',
        'children': [
            {
                'type': 'OptionBox',
                'options': ['a', 'b', 'c']
            }
        ]
    }
    return build_widget_tree(parent, item)

def demo_listbox():
    widgettree = {
        'type': 'ListBox',
        'address': 'lbox',
        'visible_entries': 5
    }
    
    app = Application(widgettree)
    for i in range(10):
        make_list_item(app.frame.get_child('lbox'), i)
    app.frame.pack()
    app.run()
